class MagicExtend:
    """
    Create a magic @extend decorator which allows to easily monkey patch
    classes of a given module. The intended usage is the following:

    # in module a.py
    from .util import MagicExtend

    extend = MagicExtend()

    @extend.register
    class Foo:
        pass

    @extend.register
    class Bar:
        pass

    # in module b.py
    import a
    @a.extend
    class Foo:
        x = 42

    @a.extend
    class Bar:
        y = 43
    """

    def __init__(self, modname):
        self.modname = modname
        self.classes = {}

    def __repr__(self):
        return f'<MagicExtend {self.modname!r}>'

    def register(self, cls):
        self.classes[cls.__name__] = cls
        return cls

    def dump_skeleton(self):
        """
        Dump a template skeleton to extend all the registered classes
        """
        lines = []
        w = lines.append
        package, mod = self.modname.rsplit('.', 1)
        w(f'from {package} import {mod}')
        w('')
        for name, cls in self.classes.items():
            w(f'@{mod}.extend')
            w(f'class {name}:')
            w(f'    pass')
            w('')
            w('')
        return '\n'.join(lines)

    def __call__(self, new_class):
        """
        This is the @extend decorator
        """
        name = new_class.__name__
        if name not in self.classes:
            raise AttributeError(f'class {name} is not registered as extendable')
        cls = self.classes[name]
        for key, value in new_class.__dict__.items():
            if key not in ('__dict__', '__doc__', '__module__', '__weakref__'):
                setattr(cls, key, value)
        return cls



"""
SPy parser.

The goal is to turn the source code into the SPy IR. For the sake of
simplicity, for now we are reusing the Python parser, so the job of this
module is to translate the Python AST into the SPy IR.  Eventually, we will
probably want to write our own SPy-specific parser, to be able to tweak the
syntax.

Note that for now the SPy IR is just an Abstract Syntax Tree. To avoid
confusion, we adopt the following naming convention:

  - When we talk about the AST, we talk about the Python AST, i.e. the ast.*
    module.  There should be no references to the Python AST outside this
    module.

  - When we talk SPy Abstract Syntax Tree, we call it IR, and it's
    implemented by spy.ir.*.
"""

import ast
from astpretty import pprint as pp
from spy.util import MagicExtend
from spy import ir
from spy.vm import wt, W_Module, W_Function


def parse(src):
    tree = ast.parse(src)
    return tree.to_spy_ir()


def parse_and_meta_eval(src):
    from spy.vm import SPyVM
    w_mod = parse(src)
    vm = SPyVM(w_mod)
    w_mod2 = vm.meta_eval_module(w_mod)
    return w_mod2

def parse_type(name: ast.Name):
    assert type(name) is ast.Name


# ===================
# AST magic
# ===================

def make_ast_extend():
    """
    some magic to let us @extend ast.* classes
    """
    ast_extend = MagicExtend('ast')
    for obj in ast.__dict__.values():
        if type(obj) is type and issubclass(obj, ast.AST):
            ast_extend.register(obj)
    return ast_extend

ast_extend = make_ast_extend()


@ast_extend
class Module:
    def to_spy_ir(self):
        for child in self.body:
            if not isinstance(child, ast.FunctionDef):
                raise Exception('Invalid child of module: %s' % child)

        funcs_w = [f.to_spy_ir() for f in self.body]
        name = 'mymodule' # XXX
        w_mod = W_Module(name)
        for w_func in funcs_w:
            w_mod.add(w_func)
        return w_mod


@ast_extend
class FunctionDef:
    def to_spy_ir(self):
        name = self.name
        param_names, args_wt = self.args.as_spy_params()
        wt_return = self.returns.as_spy_type()
        #
        ir_stmts = []
        for py_stmt in self.body:
            ir_stmts.append(py_stmt.to_spy_ir())
        ir_body = ir.Block(ir_stmts)
        wt_func = wt.function(args_wt, wt_return)
        return W_Function(name, wt_func, param_names, ir_body)


@ast_extend
class Return:
    def to_spy_ir(self):
        return ir.Return(self.value.to_spy_ir())

@ast_extend
class Name:
    def as_spy_type(self):
        try:
            return getattr(wt, self.id)
        except AttributeError:
            raise Exception(f'Unknown SPY type: {name.id}')

    def to_spy_ir(self):
        return ir.GetLocal(self.id)


@ast_extend
class arguments:
    def as_spy_params(self):
        param_names = []
        args_wt = []
        for arg in self.args:
            if arg.annotation is None:
                raise Exception('function arguments must have a type')
            args_wt.append(arg.annotation.as_spy_type())
            param_names.append(arg.arg)
        return param_names, args_wt


@ast_extend
class Expr:
    def to_spy_ir(self):
        return self.value.to_spy_ir()


@ast_extend
class BinOp:
    def to_spy_ir(self):
        op = self.op.__class__.__name__
        if op == 'Add':
            # XXX: here we assume it's i32 but we need a proper typechecking phase
            return ir.Add(self.left.to_spy_ir(),
                          self.right.to_spy_ir())
        elif op == 'Mult':
            return ir.Mul(self.left.to_spy_ir(),
                          self.right.to_spy_ir())
        else:
            raise Exception(f'Unknown op: {op}')


@ast_extend
class Constant:
    def to_spy_ir(self):
        assert type(self.value) is int
        return ir.Const(self.value)


@ast_extend
class Call:
    def to_spy_ir(self):
        # this is very fragile: it assumes that the callee exists in the
        # module that we are compiling. We should check this at compile time
        # and report errors properly. Also, we assume that we are calling a
        # Name.
        assert isinstance(self.func, ast.Name)
        callee = self.func.id
        args = [arg.to_spy_ir() for arg in self.args]
        return ir.Call(callee, args)


@ast_extend
class Compare:
    def to_spy_ir(self):
        assert len(self.ops) == 1
        assert len(self.comparators) == 1
        opname = self.ops[0].__class__.__name__ # Gt, Lt, etc.
        ir_op = getattr(ir, opname)
        ir_left = self.left.to_spy_ir()
        ir_right = self.comparators[0].to_spy_ir()
        return ir_op(ir_left, ir_right)


@ast_extend
class AnnAssign:
    def to_spy_ir(self):
        assert isinstance(self.target, ast.Name)
        name = self.target.id
        wt_type = self.annotation.as_spy_type()
        ir_expr = self.value.to_spy_ir()
        return ir.Assign(name, ir_expr, wt_type=wt_type)


@ast_extend
class Assign:
    def to_spy_ir(self):
        assert len(self.targets) == 1
        target = self.targets[0]
        assert isinstance(target, ast.Name)
        name = target.id
        ir_expr = self.value.to_spy_ir()
        return ir.Assign(name, ir_expr)



from spy import ir
from spy.vm import SPyVM
from spy.vm import wt, W_Primitive, W_Function, W_Module
from spy.tests.support import spy_parse

def make_vm():
    w_mod = W_Module('mod_for_tests')
    return SPyVM(w_mod)

def test_W_Primitive():
    w_a = W_Primitive(wt.i32, 10)
    assert w_a.unwrap() == (wt.i32, 10)

def test_wrap():
    vm = make_vm()
    w_a = vm.wrap(42)
    assert isinstance(w_a, W_Primitive)
    assert w_a.unwrap() == (wt.i32, 42)

def test_i32_const():
    expr = ir.Const(42)
    vm = None
    frame = None
    w_res = expr.vm_eval(vm, frame)
    assert isinstance(w_res, W_Primitive)
    assert w_res.unwrap() == (wt.i32, 42)

def test_i32_add():
    a = ir.Const(20)
    b = ir.Const(22)
    expr = ir.Add(a, b)
    vm = None
    frame = None
    w_res = expr.vm_eval(vm, frame)
    assert w_res.unwrap() == (wt.i32, 42)

def test_i32_mul():
    a = ir.Const(5)
    b = ir.Const(6)
    expr = ir.Mul(a, b)
    vm = None
    frame = None
    w_res = expr.vm_eval(vm, frame)
    assert w_res.unwrap() == (wt.i32, 30)

def test_block():
    expr = ir.Block([
        ir.Const(1),
        ir.Const(2),
        ir.Const(3),
    ])
    vm = make_vm()
    frame = vm.new_frame()
    w_res = expr.vm_eval(vm, frame)
    assert w_res.unwrap() == (wt.i32, 3)

def test_locals_and_block():
    expr = ir.Block([
        ir.Assign('x', ir.Const(123)),
        ir.GetLocal('x'),
    ])
    vm = make_vm()
    frame = vm.new_frame()
    w_res = expr.vm_eval(vm, frame)
    assert w_res.unwrap() == (wt.i32, 123)

def test_func():
    wt_func = wt.function([wt.i32, wt.i32], wt.i32)
    param_names = ['x', 'y']
    body = ir.Block([
        ir.Return(
            ir.Add(
                ir.GetLocal('x'),
                ir.GetLocal('y'))),
        ])
    w_func = W_Function('add', wt_func, ['x', 'y'], body)
    vm = make_vm()
    args_w = [vm.wrap(10), vm.wrap(20)]
    w_res = w_func.call(vm, args_w)
    assert w_res.unwrap() == (wt.i32, 30)

def test_call():
    mod = spy_parse("""
    def a(x: i32) -> i32:
        return b(x) * 2

    def b(x: i32) -> i32:
        return x + 3
    """)
    vm = SPyVM(mod)
    w_a = vm.get_w_function('a')
    args_w = [vm.wrap(5)]
    w_res = w_a.call(vm, args_w)
    assert w_res.unwrap() == (wt.i32, 16)



import os
import pytest
import textwrap
import sexpdata
from spy import parser
from spy.vm import SPyVM, meta_eval
from spy.backend import compile_module_to_wat
import wasmtime

def spy_parse(src):
    return parser.parse(textwrap.dedent(src))

def sexp_equal(a, b):
    """
    Compare two sexp, ignoring whitespaces
    """
    sa = sexpdata.parse(a)
    sb = sexpdata.parse(b)
    return sa == sb


class InterpModuleWrapper:

    def __init__(self, w_mod):
        self._vm = SPyVM(w_mod)
        for name, w_func in self._vm.w_mod._functions_w.items():
            func_wrapper = InterpFuncWrapper(self._vm, w_func)
            setattr(self, name, func_wrapper)

class InterpFuncWrapper:

    def __init__(self, vm, w_func):
        self._vm = vm
        self._w_func = w_func

    def __call__(self, *args):
        # *args contains python-level objs. We want to wrap them into args_w
        # *and to call the func, and unwrap the result
        args_w = [self._vm.wrap(arg) for arg in args]
        w_res = self._w_func.call(self._vm, args_w)
        t, val = w_res.unwrap()
        return val


class WasmModuleWrapper:

    def __init__(self, w_mod, wasmfile):
        store = wasmtime.Store()
        module = wasmtime.Module.from_file(store.engine, wasmfile)
        instance = wasmtime.Instance(store, module, [])
        exports = instance.exports(store)
        for name, wasm_func in exports._extern_map.items():
            wasm_func_wrapper = WasmFuncWrapper(store, wasm_func)
            setattr(self, name, wasm_func_wrapper)

class WasmFuncWrapper:

    def __init__(self, store, wasm_func):
        self._store = store
        self._wasm_func = wasm_func

    def __call__(self, *args):
        return self._wasm_func(self._store, *args)


@pytest.mark.usefixtures('init')
class SPyTest:

    @pytest.fixture(params=['interp', 'meta', 'wasm'])
    def spy_eval_strategy(self, request):
        """
        Each test is run 3 times, with different eval strategies:

          - interp: the whole module is interpreted as is

          - meta: the module is partially evaluated, and then interpreted

          - wasm: the module is partially evaluated, then compiled to wasm,
            then executed by wasmtime
        """
        return request.param

    @pytest.fixture
    def init(self, tmpdir, spy_eval_strategy):
        self.tmpdir = tmpdir
        self.spy_eval_strategy = spy_eval_strategy

    def parse(self, src):
        """
        Parse the given source code and return a w_mod
        """
        return spy_parse(src)

    def compile(self, src):
        w_mod = self.parse(src)
        if self.spy_eval_strategy == 'interp':
            return InterpModuleWrapper(w_mod)
        elif self.spy_eval_strategy == 'meta':
            w_mod2 = meta_eval(w_mod)
            return InterpModuleWrapper(w_mod2)
        elif self.spy_eval_strategy == 'wasm':
            w_mod2 = meta_eval(w_mod)
            wasmfile = self._compile_wasm(w_mod2)
            return WasmModuleWrapper(w_mod, wasmfile)
        else:
            assert False

    def _compile_wasm(self, w_mod):
        wat = compile_module_to_wat(w_mod)
        watfile = self.tmpdir.join('mymod.wat')
        wasmfile = self.tmpdir.join('mymod.wasm')
        watfile.write(wat)
        ret = os.system(f'wat2wasm {watfile} -o {wasmfile}')
        assert ret == 0
        return wasmfile



import sys
import os
import pathlib

USAGE = """
Usage:
    python -m spy compile FILE
    python -m spy dump_ir
"""

def spy_compile(filename):
    """
    XXX the following is a completely hackish and demo implementation, written
    in 5 minutes just to have something which works. Refactor me please.
    """
    import spy.parser
    import spy.vm
    from spy.backend import compile_module_to_wat
    import sexpdata

    with open(filename) as f:
        src = f.read()

    w_mod = spy.parser.parse(src)
    w_mod = spy.vm.meta_eval(w_mod)
    wat = compile_module_to_wat(w_mod)
    tmpwat = '/tmp/' + filename + '.wat'
    with open(tmpwat, 'w') as f:
        f.write(wat)

    wasm = pathlib.Path(filename).with_suffix('.wasm')
    os.system(f'wat2wasm {tmpwat} -o {wasm}')


def dump_ir():
    from spy import ir
    print(ir.extend.dump_skeleton())

def main():
    if len(sys.argv) == 3 and sys.argv[1] == 'compile':
        return spy_compile(sys.argv[2])

    if len(sys.argv) == 2 and sys.argv[1] == 'dump_ir':
        return dump_ir()

    print(USAGE)

if __name__ == '__main__':
    main()



from spy import ir
from spy.vm import SPyVM, wt, meta_eval
from spy.backend import compile_module_to_wat, compile_func_to_wat
from spy.tests.support import spy_parse, sexp_equal


class TestMetaEval:

    def meta_eval(self, src):
        w_mod = spy_parse(src)
        return meta_eval(w_mod)

    def test_simple(self):
        w_mod = self.meta_eval("""
        def foo() -> i32:
            return 1+2
        """)

        w_foo = w_mod.lookup('foo')
        assert w_foo.ir_body == ir.Block([
            ir.Return(
                ir.Add(left=ir.Const(1),
                       right=ir.Const(2))
            )
        ])

    def test_red_call(self):
        w_mod = self.meta_eval("""
        def foo() -> i32:
            return 42

        def bar() -> i32:
            return foo()
        """)
        w_bar = w_mod.lookup('bar')
        assert w_bar.ir_body == ir.Block([
            ir.Return(
                ir.Call('foo', []),
            )
        ])

    def test_green_call(self):
        w_mod = self.meta_eval("""
        def FOO() -> i32:
            return 42

        def bar() -> i32:
            return FOO()
        """)
        w_bar = w_mod.lookup('bar')
        assert w_bar.ir_body == ir.Block([
            ir.Return(
                ir.Const(42)
            )
        ])

    def test_green_call_with_arguments(self):
        w_mod = self.meta_eval("""
        def FOO(a: i32, b: i32) -> i32:
            return a+b

        def bar() -> i32:
            return FOO(10, 20)
        """)
        w_bar = w_mod.lookup('bar')
        assert w_bar.ir_body == ir.Block([
            ir.Return(
                ir.Const(30)
            )
        ])



"""
SPY Typesystem.

The classes in this module represents the types which are attached to the IR
expressions and to the runtime value used by the VM.

Naming convention: instances of subclasses of W_Type are called wt_* ("wrapped
typed"). I.e.:

    w_foo = ...
    wt_foo = vm.typeof(w_value)
"""

from dataclasses import dataclass
from spy.vm.objects import W_Root

class W_Type(W_Root):
    pass

@dataclass
class W_PrimitiveType(W_Type):
    name: str

@dataclass
class W_FunctionType(W_Type):
    args_wt: list  # List[W_Type]
    wt_return: W_Type

    @property
    def n_args(self):
        return len(self.args_wt)


class wt:
    """
    This is not a real class: just a namespace to contain prebuilt types such as i32.

    Note the exception to the naming convention: since the namespace is
    already called "wt", there is no need to repeat the wt_ prefix for the
    individual fields.  We expect the rest of the code to use e.g. wt.i32
    """

    i32 = W_PrimitiveType('i32')

    @staticmethod
    def function(args_wt, wt_return):
        return W_FunctionType(args_wt, wt_return)



from dataclasses import dataclass
from spy.vm import W_Type

class SymbolAlreadyDeclaredError(Exception):
    """
    A symbol is being redeclared
    """


@dataclass
class Symbol:
    name: str
    wt_type: W_Type


class SymTable:
    def __init__(self, scope_name, parent=None):
        self._scope_name = scope_name
        self._symbols = {}
        self._parent = parent

    def declare(self, name, wt_type):
        if name in self._symbols:
            raise SymbolAlreadyDeclaredError(name)
        self._symbols[name] = s = Symbol(name, wt_type)
        return s

    def lookup(self, name):
        if name in self._symbols:
            # found in the local scope
            return self._symbols[name]
        elif self._parent is not None:
            return self._parent.lookup(name)
        else:
            return None # not found



import operator
from spy import ir
from spy.vm import wt, W_Primitive, W_Function
from spy.vm.objects import FunctionReturn

@ir.extend
class Node:
    def vm_meta_eval(self, vm):
        """
        Perform meta evaluation and return an AST tree where some of the nodes has
        already been evaluated.

        The default implementation is to return a copy of itself after having
        recursively applied vm_meta_eval() to all children node.
        """
        def apply_vm_meta_eval_maybe(x):
            if hasattr(x, 'vm_meta_eval'):
                return x.vm_meta_eval(vm)
            return x

        new_fields = {}
        for name in self.__dataclass_fields__:
            value = getattr(self, name)
            if isinstance(value, ir.Node):
                # recursively apply vm_meta_eval
                value = value.vm_meta_eval(vm)
            elif isinstance(value, list):
                value = [apply_vm_meta_eval_maybe(x) for x in value]
            #
            new_fields[name] = value
        return self.__class__(**new_fields)


@ir.extend
class Expr:
    def vm_eval(self, vm, frame):
        raise NotImplementedError


@ir.extend
class Const:
    def vm_eval(self, vm, frame):
        return W_Primitive(wt.i32, self.value)


@ir.extend
class BinOp:

    def _binop_eval(self, vm, frame, op):
        w_left = self.left.vm_eval(vm, frame)
        w_right = self.right.vm_eval(vm, frame)
        # XXX we need proper typecheck and better error reporting
        assert isinstance(w_left, W_Primitive)
        assert isinstance(w_right, W_Primitive)
        res = op(w_left.value, w_right.value)
        return W_Primitive(wt.i32, res)


@ir.extend
class Add:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.add)

@ir.extend
class Mul:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.mul)

@ir.extend
class Lt:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.lt)

@ir.extend
class LtE:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.le)

@ir.extend
class Gt:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.gt)

@ir.extend
class GtE:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.ge)

@ir.extend
class Eq:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.eq)

@ir.extend
class NotEq:
    def vm_eval(self, vm, frame):
        return self._binop_eval(vm, frame, operator.ne)


@ir.extend
class Block:

    def vm_eval(self, vm, frame):
        val = None
        for s in self.statements:
            val = s.vm_eval(vm, frame)
        return val


@ir.extend
class GetLocal:

    def vm_eval(self, vm, frame):
        return frame.get(self.name)


@ir.extend
class Assign:

    def vm_eval(self, vm, frame):
        val = self.expr.vm_eval(vm, frame)
        frame.set(self.name, val)


@ir.extend
class Return:

    def vm_eval(self, vm, frame):
        w_result = self.expr.vm_eval(vm, frame)
        raise FunctionReturn(w_result)


@ir.extend
class Call:

    def vm_eval(self, vm, frame):
        # XXX better error reporting in case self.func doesn't exist
        w_func = vm.get_w_function(self.func)
        assert isinstance(w_func, W_Function)
        args_w = [arg.vm_eval(vm, frame) for arg in self.args]
        return w_func.call(vm, args_w)

    def vm_meta_eval(self, vm):
        w_func = vm.get_w_function(self.func)
        if w_func.is_green():
            args_w = []
            assert w_func.wt_func.n_args == len(self.args) # XXX better error reporting
            for arg in self.args:
                assert arg.is_green()
                # XXX this is wrong, we should pass a frame and/or have
                # another way to go from i32_const to W_Primitive
                w_arg = arg.vm_eval(vm, None)
                args_w.append(w_arg)

            w_result = w_func.call(vm, args_w)
            # XXX for now we support only i32 prebuilt constants
            assert isinstance(w_result, W_Primitive)
            assert w_result.wt_value == wt.i32
            return ir.Const(w_result.value)
        else:
            args = [arg.vm_meta_eval(vm) for arg in self.args]
            return ir.Call(self.func, args)



from spy.vm.objects import W_Root
from spy.vm import wt


class W_Primitive(W_Root):

    def __init__(self, wt_value, value):
        # we should add proper error reporting, but for now it's fine
        assert wt_value == wt.i32
        assert isinstance(value, int)
        self.wt_value = wt_value
        self.value = value

    def unwrap(self):
        return self.wt_value, self.value



from dataclasses import dataclass
from spy.util import MagicExtend
from spy.vm import wt, W_Type

extend = MagicExtend('spy.ir')

def irclass(cls):
    """
    This is just a shortcut: it automatically applies the following decorators
    to all IR nodes:

        @extend.register
        @dataclass
    """
    return extend.register(dataclass(cls))


# ==========================
# AST hierarchy
# ==========================

@irclass
class Node:

    def is_green(self):
        return False


@irclass
class Expr(Node):

    def get_type(self):
        raise NotImplementedError


@irclass
class Const(Expr):
    # XXX for now we assume that it's an i32 const
    value: int

    def __init__(self, value):
        self.value = value

    def get_type(self):
        return wt.i32

    def is_green(self):
        return True


@irclass
class BinOp(Expr):
    left: Expr
    right: Expr


@irclass
class Add(BinOp):

    def get_type(self):
        return wt.i32


@irclass
class Mul(BinOp):

    def get_type(self):
        return wt.i32


@irclass
class Block(Expr):
    statements: list # List[Expr]


@irclass
class GetLocal(Expr):
    name: str


@irclass
class Assign(Expr):
    name: str
    expr: Expr
    wt_type: W_Type = None


@irclass
class Call(Expr):
    func: str
    args: list  # List[Expr]


@irclass
class Lt(BinOp):
    pass

@irclass
class LtE(BinOp):
    pass

@irclass
class Gt(BinOp):
    pass

@irclass
class GtE(BinOp):
    pass

@irclass
class Eq(BinOp):
    pass

@irclass
class NotEq(BinOp):
    pass


@irclass
class Return(Node):
    expr: Expr



import pytest
import textwrap
from spy.util import MagicExtend

def test_MagicExtend():

    extend = MagicExtend('MyModule')

    class MyModule:
        @extend.register
        class A:
            pass

    @extend
    class A:
        x = 42

    assert A is MyModule.A
    assert A.x == 42

    with pytest.raises(AttributeError, match='class B is not registered as extendable'):
        @extend
        class B:
            pass

def test_MagicExtend_dump_skeleton():

    extend = MagicExtend('spy.mymod')

    @extend.register
    class A:
        pass

    @extend.register
    class B:
        pass

    skeleton = extend.dump_skeleton()
    assert skeleton.strip() == textwrap.dedent("""
        from spy import mymod

        @mymod.extend
        class A:
            pass


        @mymod.extend
        class B:
            pass
    """).strip()



import pytest
from spy.symtable import SymTable, SymbolAlreadyDeclaredError
from spy.vm import wt

def test_basic():
    t = SymTable('<globals>')
    sym = t.declare('a', wt.i32)
    assert sym.name == 'a'
    assert sym.wt_type == wt.i32
    assert t.lookup('a') is sym
    assert t.lookup('I-dont-exist') is None
    #
    with pytest.raises(SymbolAlreadyDeclaredError):
        t.declare('a', wt.i32)


def test_nested_table_lookup():
    glob = SymTable('<globals>')
    loc = SymTable('loc', glob)
    sym_a = glob.declare('a', wt.i32)
    sym_b = loc.declare('b', wt.i32)
    #
    assert glob.lookup('a') is sym_a
    assert glob.lookup('b') is None
    #
    assert loc.lookup('a') is sym_a
    assert loc.lookup('b') is sym_b



from spy.tests.support import SPyTest

class TestBasic(SPyTest):

    def test_simple(self):
        mod = self.compile("""
            def foo() -> i32:
                return 42

            def bar() -> i32:
                return 43
        """)
        assert mod.foo() == 42
        assert mod.bar() == 43

    def test_args(self):
        mod = self.compile("""
            def add(x: i32, y: i32) -> i32:
                return x + y
        """)
        assert mod.add(40, 2) == 42

    def test_cmp_operators(self):
        # XXX: comparison operators should return bool, but we haven't introduced this type yet
        mod = self.compile("""
            def lt(x: i32, y: i32) -> i32: return x < y
            def le(x: i32, y: i32) -> i32: return x <= y
            def ge(x: i32, y: i32) -> i32: return x >= y
            def gt(x: i32, y: i32) -> i32: return x > y
            def eq(x: i32, y: i32) -> i32: return x == y
            def ne(x: i32, y: i32) -> i32: return x != y
        """)
        assert mod.lt(3, 4)
        assert not mod.lt(4, 3)
        assert mod.le(3, 3)
        assert not mod.le(4, 3)
        assert mod.ge(3, 3)
        assert not mod.ge(3, 4)
        assert mod.gt(4, 3)
        assert not mod.gt(3, 3)
        assert mod.eq(3, 3)
        assert not mod.eq(3, 4)
        assert mod.ne(3, 4)
        assert not mod.ne(3, 3)

    def test_local_set(self):
        mod = self.compile("""
            def foo() -> i32:
                a: i32 = 42
                return a

            def bar() -> i32:
                a: i32 = 42
                a = 100
                return a
        """)
        assert mod.foo() == 42
        assert mod.bar() == 100



"""
Very simple VM to execute the SPY IR.
"""

from dataclasses import dataclass
from spy import ir
from spy.vm.typesystem import wt, W_Type
from spy.vm.objects import W_Root, W_Function, W_Module
from spy.vm.primitive import W_Primitive

def meta_eval(w_mod):
    """
    Perform the meta-eval phase on the given module.
    """
    vm = SPyVM(w_mod)
    return vm.meta_eval_module(w_mod)


class Frame:

    def __init__(self):
        self.locals = {}

    def set(self, name, value):
        assert isinstance(value, W_Root)
        self.locals[name] = value

    def get(self, name):
        return self.locals[name]


class SPyVM:
    """
    A Virtual Machine to execute SPy code.
    """

    def __init__(self, w_mod):
        # XXX: here we are assuming that there is only a single module inside
        # a VM. This is fine for now, but we need a better model.
        self.w_mod = w_mod

    def wrap(self, value):
        """
        Useful for tests: magic funtion which wraps the given value into the most
        appropriate W_* object.
        """
        # only i32 is supported for now
        assert isinstance(value, int)
        return W_Primitive(wt.i32, value)

    def new_frame(self):
        return Frame()

    def get_w_function(self, name):
        return self.w_mod.lookup(name)

    def meta_eval_module(self, w_mod):
        """
        Run the meta-time parts of w_mod, and return another partially-evaluated w_mod
        """
        funcs_w = w_mod._functions_w.values()
        funcs_w2 = [w_f.vm_meta_eval(self) for w_f in funcs_w]
        w_mod2 = W_Module(w_mod._name)
        for w_func in funcs_w2:
            w_mod2.add(w_func)
        return w_mod2



"""
SPY IR backend.

Compiles the IR into various outputs such as WAT.
"""

import sexpdata
from spy import ir

def compile_module_to_wat(w_mod):
    """
    Compile a W_Module to WASM.
    """
    funcs_w = w_mod._functions_w.values()
    funcs = [compile_func_to_sexp(w_f) for w_f in funcs_w]
    exports = [[Sym('export'), w_f.name, [Sym('func'), Name(w_f.name)]]
               for w_f in funcs_w]
    sexp = [Sym('module')] + funcs + exports
    return sexpdata.dumps(sexp)

def compile_func_to_sexp(w_func):
    sexp = [Sym('func'), Name(w_func.name)]
    for param, wt_param in w_func.get_params():
        sexp.append([Sym('param'), Name(param), Sym(wt_param.name)])
    sexp.append([Sym('result'), Sym(w_func.wt_func.wt_return.name)])
    assert isinstance(w_func.ir_body, ir.Block)

    # first, we declare all local variables. XXX: for now we simple search for
    # the appropriate ir.local_set statements, but we soon need a proper
    # symbol table and typechecking phase.
    for ir_stmt in w_func.ir_body.statements:
        if isinstance(ir_stmt, ir.Assign) and ir_stmt.wt_type is not None:
            sexp.append([Sym('local'), Name(ir_stmt.name,), Sym(ir_stmt.wt_type.name)])

    for ir_stmt in w_func.ir_body.statements:
        sexp.append(ir_stmt.to_sexp())
    return sexp

def compile_func_to_wat(w_func):
    sexp = compile_func_to_sexp(w_func)
    return sexpdata.dumps(sexp)


class Sym(sexpdata.SExpBase):
    """
    Similar to sexpdata.Symbol, but doesn't quote its value
    """

    def tosexp(self, tosexp=None):
        return self._val

class Name(sexpdata.SExpBase):
    """
    Like Sym, but automatically add a $ prefix
    """

    def tosexp(self, tosexp=None):
        return '$' + self._val


@ir.extend
class Expr:
    def to_wat(self):
        return sexpdata.dumps(self.to_sexp())

    def to_sexp(self):
        raise NotImplementedError(self.__class__.__name__)


@ir.extend
class Const:
    def to_sexp(self):
        return [Sym('i32.const'), self.value]


@ir.extend
class Add:
    def to_sexp(self):
        return [Sym('i32.add'),
                self.left.to_sexp(),
                self.right.to_sexp()]


@ir.extend
class Lt:
    def to_sexp(self):
        # XXX: this assumes that the operands are i32, but we need a better
        # typechecking
        return [Sym('i32.lt_s'),
                self.left.to_sexp(),
                self.right.to_sexp()]


@ir.extend
class LtE:
    def to_sexp(self):
        # XXX: this assumes that the operands are i32, but we need a better
        # typechecking
        return [Sym('i32.le_s'),
                self.left.to_sexp(),
                self.right.to_sexp()]


@ir.extend
class GtE:
    def to_sexp(self):
        # XXX: this assumes that the operands are i32, but we need a better
        # typechecking
        return [Sym('i32.ge_s'),
                self.left.to_sexp(),
                self.right.to_sexp()]

@ir.extend
class Gt:
    def to_sexp(self):
        # XXX: this assumes that the operands are i32, but we need a better
        # typechecking
        return [Sym('i32.gt_s'),
                self.left.to_sexp(),
                self.right.to_sexp()]


@ir.extend
class Eq:
    def to_sexp(self):
        # XXX: this assumes that the operands are i32, but we need a better
        # typechecking
        return [Sym('i32.eq'),
                self.left.to_sexp(),
                self.right.to_sexp()]


@ir.extend
class NotEq:
    def to_sexp(self):
        # XXX: this assumes that the operands are i32, but we need a better
        # typechecking
        return [Sym('i32.ne'),
                self.left.to_sexp(),
                self.right.to_sexp()]


@ir.extend
class Block:
    def to_sexp(self):
        res = [Sym('block')]
        for child in self.statements:
            res.append(child.to_sexp())
        return res


@ir.extend
class GetLocal:
    def to_sexp(self):
        return [Sym('local.get'), Name(self.name)]


@ir.extend
class Assign:
    def to_sexp(self):
        return [Sym('local.set'),
                Name(self.name),
                self.expr.to_sexp()]


@ir.extend
class Call:

    def to_sexp(self):
        res = [Sym('call'), Name(self.func)]
        for arg in self.args:
            res.append(arg.to_sexp())
        return res


@ir.extend
class Return:

    def to_sexp(self):
        return [Sym('return'), self.expr.to_sexp()]



import pytest
from spy import ir
from spy.vm import wt
from spy.tests.support import SPyTest

class TestParser(SPyTest):

    @pytest.fixture
    def spy_eval_strategy(self, request):
        return 'this-is-only-for-parsing'

    def test_minimal(self):
        w_mod = self.parse("""
        def foo() -> i32:
            return 42
        """)
        w_func = w_mod.lookup('foo')
        assert w_func.name == 'foo'
        assert w_func.wt_func == wt.function([], wt.i32)
        assert w_func.param_names == []
        #
        ir_expected = ir.Block([
            ir.Return(
                ir.Const(42)
            ),
        ])
        assert w_func.ir_body == ir_expected

    def test_add(self):
        w_mod = self.parse("""
        def add(x: i32, y: i32) -> i32:
            return x+y
        """)
        w_func = w_mod.lookup('add')
        assert w_func.name == 'add'
        assert w_func.wt_func == wt.function([wt.i32, wt.i32], wt.i32)
        assert w_func.param_names == ['x', 'y']
        ir_expected = ir.Block([
            ir.Return(
                ir.Add(
                    left=ir.GetLocal('x'),
                    right=ir.GetLocal('y')
                )
            )
        ])
        assert w_func.ir_body == ir_expected



from setuptools import setup, find_packages

setup(name="spylang",
      author='Antonio Cuni',
      author_email='anto.cuni@gmail.com',
      url='https://github.com/antocuni/spylang',
      packages=find_packages(),
      zip_safe=False,
      )



from dataclasses import dataclass

class W_Root:
    """
    Base class for all SPy VM objects.
    """

class W_Module(W_Root):
    """
    A module is the basic unit of compilation.

    A module represents a namespace which contains functions and other
    objects. It is compiled into a WASM module.
    """

    def __init__(self, name):
        self._name = name
        self._functions_w = {}

    def add(self, w_func):
        # XXX we need to think and decide what happens in this case
        assert w_func.name not in self._functions_w
        self._functions_w[w_func.name] = w_func

    def lookup(self, func_name):
        # XXX which kind of exception do we want to raise here? An
        # interp-level or app-level exception?
        return self._functions_w[func_name]


class FunctionReturn(Exception):
    """
    Not a real exception. Used to implement the 'return' statement.
    """

    def __init__(self, w_result):
        self.w_result = w_result


@dataclass
class W_Function(W_Root):

    name: str
    wt_func: "W_Type"
    param_names: list  # List[str]
    ir_body: "ir.expr"

    def post_init(self):
        assert len(wt_func.args_wt) == len(self.param_names)

    def get_params(self):
        """
        Return a list of [(param_name, wt_type), ...]
        """
        return zip(self.param_names, self.wt_func.args_wt)

    def __str__(self):
        return '<W_Function %s>' % (self.name)

    def call(self, vm, args_w):
        """
        Create a new frame, set the arguments, eval the body
        """
        # XXX: we need better typecheck and error reporting
        assert len(args_w) == self.wt_func.n_args
        frame = vm.new_frame()
        for w_arg, wt_arg, name in zip(args_w, self.wt_func.args_wt, self.param_names):
            assert w_arg.wt_value == wt_arg # poor's man typecheck
            frame.set(name, w_arg)
        try:
            w_result = self.ir_body.vm_eval(vm, frame)
            # a function body should never return a value. If it does, it's
            # probably a bug in the parser
            assert w_result is None
        except FunctionReturn as exc:
            return exc.w_result

    def is_green(self):
        # XXX: for now the green functions are the ones in UPPERCASE. We might
        # want to switch to a dedicated syntax at some point.
        return self.name.isupper()

    def vm_meta_eval(self, vm):
        ir_body = self.ir_body.vm_meta_eval(vm)
        return W_Function(self.name, self.wt_func, self.param_names, ir_body)
